/* Copyright (c) 1997, MpegTV LLC */

#ifndef lint
static char rcsid[] = "$Header: /net/hopkins/var/home/tristan/mpeg/dec/RCS/ci_file.c,v 1.2 1997/08/21 11:51:26 tristan Exp $";
static char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

#include "mtvp/misc.h"

#include "mtvp/sih_api.h"

/*
 * This is a working example of SIH plug-in.
 * See the README file.
 */

/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/
/*
 * This SIH plugin only recognizes URL's starting with the MY_PROTOCOL string.
 */
 
/* In this example, we use "file:" to override the built-in "file:" handler */

#define MY_PROTOCOL "file:"		
#define MY_PROTOCOL_SIZE (sizeof MY_PROTOCOL - 1)

/*--------------------------------------------------------------------------+
| Data types																|
+--------------------------------------------------------------------------*/
/* handler data */
typedef struct {
	const SIH_RESOURCES *sih_resources;
} HANDLER_DATA;

/* stream data */
typedef struct {
	HANDLER_DATA *handler_data;	/* should always be there */
	int		fd;
} STREAM_DATA;

/*
 * In this example, our STREAM_DATA structure contains only a file descriptor,
 * in addition to the pointer to the HANDLER_DATA (which should always be there).
 * The STREAM_DATA structure can contain absolutely anything you want,  The nature of
 * this structure is only known to the SIH routines defined in the SIH plug-in.
 */


/*--------------------------------------------------------------------------+
| _init																		|
+--------------------------------------------------------------------------*/
int
_init()
{
	/*
	 * If the library exports a routine named  _init(),  then  that
     * code  is  executed  before  dlopen()  returns.
     */

	printf("_init(): loading custom SIH handler\n");
		
	return 0;
}


/*--------------------------------------------------------------------------+
| _fini																		|
+--------------------------------------------------------------------------*/
int
_fini()
{
	/*
	 * If the dynamic library exports a routine named _fini(),
	 * then that routine is called just before the library is unloaded.
	 */

	printf("_fini(): loading custom SIH handler\n");
		
	return 0;
}

/*--------------------------------------------------------------------------+
| SIH_init																	|
+--------------------------------------------------------------------------*/
static void *
SIH_init(
	const SIH_RESOURCES *sih_resources,
	const char *p
)
{
	HANDLER_DATA *handler_data;

	printf("SIH_init with \"%s\"\n", p);

	/* allocate storage for our opaque HANDLER_DATA structure */
	handler_data = calloc(1, sizeof(HANDLER_DATA));
	
	if (handler_data == NULL) {
		perror("SIH_init");
		return NULL;
	}

	handler_data->sih_resources = sih_resources;
	
	/*
	 * Return a pointer to our HANDLER_DATA structure.
	 * Note that handler_init is defined as returning a (void *) pointer,
	 * because the returned pointer is opaque, i.e. HANDLER_DATA is unknown
	 * from the caller.
	 */
	return (void *)handler_data;
}


/*--------------------------------------------------------------------------+
| SIH_probe_url															|
+--------------------------------------------------------------------------*/
static int
SIH_probe_url(
	const SIH_RESOURCES *sih_resources,
	const char *url
)
{	
	if (strncmp(MY_PROTOCOL, url, MY_PROTOCOL_SIZE) != 0) {
		printf("SIH_probe_url: URL %s cannot be opened by this handler\n", url);
		
		errno = EINVAL;
		return -1;
	}

	printf("SIH_probe_url: URL %s can be opened by this handler\n", url);
	
	return 0;
}

/*--------------------------------------------------------------------------+
| SIH_open_url															|
+--------------------------------------------------------------------------*/
static void *
SIH_open_url(
	void *void_handler_data,		/* pointer that was returned by SIH_init */
	const char *url	/* assumed to be a file:filename in this example */
)
{
	HANDLER_DATA *handler_data = (HANDLER_DATA *)void_handler_data;
	STREAM_DATA *stream_data;
	const char *filename;
	int		fd;

	/*
	 * The parameter of SIH_open_url is an URL passed by the application
	 * to the COM_SIH_OPEN_URL command.  The URL identifies the stream
	 * to be opened.
	 *
	 * SIH_open_url returns a pointer to an opaque object that
	 * identifies the stream, i.e.
	 * only your SIH plug-in known about the structure of the
	 * opaque object.  If the specified stream cannot be opened,
	 * SIH_open_url returns NULL and sets errno.
	 */	

	filename = url + strlen(MY_PROTOCOL);
	
	printf("SIH_open_url calls open(\"%s\", O_RDONLY)\n", filename);
	
	/* check that sih_feedback() works */
	(*handler_data->sih_resources->feedback)(handler_data->sih_resources, url, strlen(url) + 1);
	
	/* in this example, we just open the specified file */
	fd = open(filename, O_RDONLY);
	
	if (fd == -1) {
		perror("SIH_open_url");
		return NULL;
	}
	
	/* allocate storage for our opaque STREAM_DATA structure */
	stream_data = calloc(1, sizeof(STREAM_DATA));
	
	if (stream_data == NULL) {
		perror("SIH_open_url");
		return NULL;
	}
	
	stream_data->handler_data = handler_data;

	/* the file descriptor for the opened file is stored in our STREAM_DATA structure */
	stream_data->fd = fd;
        	
	/*
	 * Return a pointer to our PLUGIN_DATA structure.
	 * Note that SIH_open_url is defined as returning a (void *) pointer,
	 * because the returned pointer is opaque, i.e. STREAM_DATA is unknown
	 * from the caller (mtvp).
	 */
	return stream_data;
}

/*--------------------------------------------------------------------------+
| SIH_fd																|
+--------------------------------------------------------------------------*/
static int
SIH_fd(
	void *void_stream_data
)
{
	STREAM_DATA *stream_data = (STREAM_DATA *)void_stream_data;

	/*
	 * Returns the file descriptor associated with the stream.
	 * 
	 * This routine must return a real file descriptor.
	 * The Player calls a blocking select on this file descriptor
	 * and if there is something available for reading on it, calls SIH_read
	 * (provided that there is buffer space available to read some input data).
	 * 
	 * After calling SIH_fd, mtvp calls fcntl with F_SETFL
	 * to set O_NONBLOCK on the fd.
	 */

	printf("SIH_fd returns %d\n", stream_data->fd);
	
	return stream_data->fd;
}


/*--------------------------------------------------------------------------+
| SIH_close																|
+--------------------------------------------------------------------------*/
static int
SIH_close(
	void *void_stream_data
)
{
	STREAM_DATA *stream_data = (STREAM_DATA *)void_stream_data;

	/*
	 * Close the stream.
	 * 
	 * This routine must close every fd opened by SIH_open_url and free the PLUGIN_DATA
	 * structure.  All the resources allocated for the stream should be released.
	 * 
	 * After a call to SIH_close, the "void *stream"
	 * should be considered invalid.
	 */

	printf("SIH_close\n");
	
	close(stream_data->fd);
	
	free(stream_data);
	
	return 0;
}

/*--------------------------------------------------------------------------+
| SIH_free																|
+--------------------------------------------------------------------------*/
static int
SIH_free(
	void *void_handler_data
)
{
	HANDLER_DATA *handler_data = (HANDLER_DATA *)void_handler_data;

	printf("SIH_free\n");

	free(handler_data);
	
	return 0;
}


/*--------------------------------------------------------------------------+
| SIH_read																	|
+--------------------------------------------------------------------------*/
static int
SIH_read(
	void *void_stream_data,
	char *buf,
	size_t count
)
{
	STREAM_DATA *stream_data = (STREAM_DATA *)void_stream_data;
	int		n;

	/*
	 * SIH_read must behave exactly like the
	 * regular read routine when used on a fd
	 * that has O_NONBLOCK set.
	 * 
	 * If data becomes available for reading on the fd returned by SIH_fd,
	 * SIH_read is called. Note that mtvp may call SIH_read at any time, even
	 * when there is no data available on the fd.
	 * 
	 * SIH_read should always behave like a non-blocking read
	 * (i.e. like read behaves when the flag O_NONBLOCK set).
	 * This means that SIH_read should never block, even
	 * if there is not enough data available.
	 * 
	 * If there is no data available,
	 * SIH_read should return -1 with errno set to EAGAIN.
	 * 
	 * If there is less data available than requested,
	 * SIH_read should copy all the data available into the
	 * buffer and return the number of bytes copied.
	 * 
	 * SIH_read returning 0 means EOF.
	 * 
	 * SIH_read returning -1 with errno NOT set to EAGAIN
	 * means an error occured.  All read errors (except EAGAIN)
	 * are considered like EOF.
	 */

	n = read(stream_data->fd, buf, count);
	
	printf("SIH_read calls read(%d, buf, %d), returns %d\n", stream_data->fd, count, n);

#if 0
	/* uncomment this to check that feedback() works in SIH_read */
	{
		const char buf[] = "calling feedback in SIH_read";
		(*stream_data->handler_data->sih_resources->feedback)(stream_data->handler_data->sih_resources, buf, sizeof buf);
	}
#endif
	
	return n;
}



/*--------------------------------------------------------------------------+
| SIH_lseek																	|
+--------------------------------------------------------------------------*/
static off_t
SIH_lseek(
	void *void_stream_data,
	off_t offset,
	int whence
)
{
	STREAM_DATA *stream_data = (STREAM_DATA *)void_stream_data;
	off_t	n;
	
	/*
	 * SIH_lseek must behave exactly like the regular lseek routine.
	 * 
	 * In particular:
	 * 
	 * SIH_lseek should return -1 with errno set to ESPIPE
	 * if the input does not support seeking.
	 */
	 
	n = lseek(stream_data->fd, offset, whence);

	printf("SIH_lseek calls lseek(%d, %ldL, %d), returns %ld\n", stream_data->fd, (long)offset, whence, (long)n);
	
	return n;
}


/*--------------------------------------------------------------------------+
| SIH_size																	|
+--------------------------------------------------------------------------*/
static off_t
SIH_size(
	void *void_stream_data
)
{
	STREAM_DATA *stream_data = (STREAM_DATA *)void_stream_data;
	struct	stat stat_buf;
	off_t	size;
	int		n;

	/*
	 * SIH_size, if defined, must return the size of the file, in bytes.
	 * 
	 * If SIH_size is not defined, appropriate calls to SIH_lseek are used
	 * to determine the size of the file.
	 */
	 
	n = fstat(stream_data->fd, &stat_buf);

	printf("SIH_size calls fstat(%d, ...), returns %d, size = %lu\n", stream_data->fd, n, (u_long)stat_buf.st_size);
	
	if (n != 0) {
		return (off_t)-1;
	}
	
	return stat_buf.st_size;
}


/*--------------------------------------------------------------------------+
| SIH_cntl																|
+--------------------------------------------------------------------------*/
static int
SIH_cntl(
	void *void_handler_stream,
	const void *data,
	u_int		size
)
{
	/*
	 * SIH_cntl is called if and only if the application sends
	 * the message COM_SIH_CNTL.  The parameter "data" is a pointer to
	 * an arbitrary chunk of data specified by the application.
	 * "size" is the size of the chunk of data.
	 */
	 
	/*
	 * In this example, the data block passed to SIH_cntl() is known to be
	 * a string, so we just print it to show that we received it.
	 */
	 
	printf("SIH_cntl: data = \"%s\", size = %u\n", (const char *)data, size);
	
	return 0;
}

/*--------------------------------------------------------------------------+
| SIH_register																|
+--------------------------------------------------------------------------*/
int
SIH_register(SIH_EXPOSED *exposed, int version)
{
	if (version == SIH_VERSION) {

	 	exposed->SIH_version = SIH_VERSION;
		exposed->SIH_init = SIH_init;
		exposed->SIH_free = SIH_free;
		exposed->SIH_probe_url = SIH_probe_url;
		exposed->SIH_open_url = SIH_open_url;
		exposed->SIH_close = SIH_close;
		exposed->SIH_read = SIH_read;
		exposed->SIH_lseek = SIH_lseek;
		exposed->SIH_cntl = SIH_cntl;
		exposed->SIH_fd = SIH_fd;
		exposed->SIH_size = SIH_size;
		exposed->SIH_buffer = NULL;
	}
	return SIH_VERSION;
}

/* eof */
